// handlers for the watchers, to check changes on S4, the
import fs from "fs"
import dotted from "dot-object"

import { GPS, IOS, System, Network} from "syrus4-nodejs"
import { Accelerometer, ECU, Geofences, Windows} from "syrus4-nodejs"
import { Bluetooth } from "syrus4-nodejs"
import { IButton, RFID, Temperature, Technoton, Logrotate, SafeEngine, PeopleCounter } from "syrus4-nodejs"
import { Serial } from "syrus4-nodejs"
import { Fatigue } from "syrus4-nodejs"

import dayjs from "../utils/dayjs.js"
import { syrusjs_variables, fireInternalSignal } from "./globals.js"

// sets all keys on an object as null
function clear_object(object){
	for (const key in object) {
		object[key] = null
	}
}
function isObject(object){
	return Object.prototype.toString.call(object) === '[object Object]'
}
// TODO: don't initialize all watchers unless required,
// this is a sideffect from favoring TAIP, that requires certain watchers always running
export function startWatchers(configuration) {
	function onGpsData(gps) {
		Object.assign(syrusjs_variables, dotted.dot({ $gnss: gps.coords }));
		syrusjs_variables["$gnss.heading"] = gps.coords.bearing;
		syrusjs_variables["$gnss.mph"] = gps.coords.speed * 2.23694;
		syrusjs_variables["$gnss.kph"] = gps.coords.speed * 3.6;
		syrusjs_variables["$gnss.timestamp"] = gps.timestamp;

		syrusjs_variables["$speed"] = Math.round(gps.coords.speed);
		syrusjs_variables["$mph"] = Math.round(gps.coords.speed * 2.23694);
		syrusjs_variables["$kph"] = Math.round(gps.coords.speed * 3.6);

		let time = dayjs.unix(gps.timestamp)
		syrusjs_variables["$timestamp"] = time.locale("en").format("LLLL");
		syrusjs_variables["$timestamp.es"] = time.locale("es").format("LLLL");
		syrusjs_variables["$timestamp.fr"] = time.locale("fr").format("LLLL");
		syrusjs_variables["$timestamp.pt"] = time.locale("pt").format("LLLL");
		Object.assign(syrusjs_variables, dotted.dot({ $gnss: gps.extras }));
	}
	function onGeofenceChange(fences) {
		if (!Array.isArray(fences)) {
			fences = [fences];
		}

		for (const fence of fences) {
			syrusjs_variables[`$geofences.${fence.name}.inside`] = fence.is_inside;
			syrusjs_variables[`$geofences.${fence.name}.time`] = fence.time || fence.timestamp;

			syrusjs_variables[`@last_geofence.name`] = fence.name;
			syrusjs_variables[`@last_geofence.state`] = fence.is_inside ? "inside" : "outside";
			syrusjs_variables[`@last_geofence.time`] = fence.time || fence.timestamp;
			fireInternalSignal(`@geofence.${fence.is_inside ? "enter" : "exit"}`);
		}
	}
	function onGeofenceGroupChange(groups) {
		if (!Array.isArray(groups)) {
			groups = [groups];
		}

		for (const group of groups) {
			syrusjs_variables[`$geofences.$groups.${group.name}.inside`] = group.is_inside;
			syrusjs_variables[`$geofences.$groups.${group.name}.time`] = group.time || group.timestamp;

			syrusjs_variables[`@last_geofence.group.name`] = group.name;
			syrusjs_variables[`@last_geofence.group.state`] = group.is_inside ? "inside" : "outside";
			syrusjs_variables[`@last_geofence.group.time`] = group.time;
			fireInternalSignal(`@geofence.group.${group.is_inside ? "enter" : "exit"}`);
		}
	}
	function onNetworkChange() {
		Network.getNetworks()
			.then(networks => {
				for (const net in networks) {
					const network = networks[net];
					var netname = net;
					switch (net) {
						case "wlan0":
							netname = "wifi";
							break;
						case "eth0":
							netname = "eth";
							break;
						case "ppp0":
							netname = "cell";
							break;
						default:
							break;
					}
					for (const key in network) {
						syrusjs_variables[`$net_${netname}.${key}`] = network[`${key}`];
					}
				}
			})
			.catch(console.error);
	}
	function onInputChange(ios) {
		//console.log('ios_change', ios)
		const signals = []
		for (const io in ios) {
			const io_lc = io.toLowerCase()
			const value = ios[io]
			if (io == "BAT") {
				onBatteryChange(Number(value))
			} else {
				syrusjs_variables[`$io.${io_lc}`] = value
				if (value) signals.push(`@io.${io_lc}`)
			}
		}
		fireInternalSignal(...signals)
	}
	function onBatteryChange(vdc) {
		/**
		 * battery level between 4000 and 3500
		 * keys: "$battery.level" , "$battery.voltaje"
		 * bat full 4000 >= 100%
		 * bat low  3620 <= 20%
		 * bat cri  3560 <= 10%
		 */
		if (vdc == null) {
			syrusjs_variables[`$battery.connected`] = false;
			syrusjs_variables[`$battery.voltage`] = 0;
			syrusjs_variables[`$battery.mv`] = 0;
			syrusjs_variables[`$battery.level`] = 0;
			return;
		}
		syrusjs_variables[`$battery.connected`] = true;
		syrusjs_variables[`$battery.voltage`] = vdc;
		syrusjs_variables[`$battery.mv`] = Math.round(vdc * 1000);
		var level = Math.round(((vdc * 1000 - 3500) / 500) * 100);
		if (level > 100) level = 100;
		if (level < 0) level = 0;
		if (syrusjs_variables[`$battery.level`] < 100 && level >= 100) {
			fireInternalSignal("@battery.full");
		}
		if (syrusjs_variables[`$battery.level`] > 20 && level <= 20) {
			fireInternalSignal("@battery.low");
		}
		if (syrusjs_variables[`$battery.level`] > 10 && level <= 10) {
			fireInternalSignal("@battery.critical");
		}
		syrusjs_variables[`$battery.level`] = level;
	}
	function onMotionChange(isMoving) {
		syrusjs_variables["$accelerometer.motion"] = isMoving;
	}
	function onMotionEvent(event, value, time = null) {
		syrusjs_variables[`$accelerometer.${event}.time`] = time ? time : new Date().toISOString();
		syrusjs_variables[`$accelerometer.${event}.value`] = value;
		fireInternalSignal(`@accelerometer.${event}.signal`);
	}
	const tires_dotted = {}
	function onEcuParamsChange(ecu_params) {
		const flat_data = {}
		const signals = []
		for (const key in ecu_params) {
			let value = ecu_params[key]
			if (key.startsWith('@')){
				signals.push(key)
				continue
			}
			if (key.startsWith('$')){
				flat_data[key] = value
			}
			if (isObject(value)){
				let obj = {}
				obj[key] = value
				let flat = dotted.dot(obj)
				for(const fkey in flat){
					syrusjs_variables[`$ecu.${fkey}`] = flat[fkey]
				}
				continue
			}
			syrusjs_variables[`$ecu.${key}`] = value
		}
		const sub_data = dotted.object(flat_data)
		const tires = sub_data['$tires']
		if(tires){
			const all_tires = []
			const tire_signals = {
				"alarm_ok": [],
				"sensor_mute": [],
				"sensor_defective": [],
				// "sensor_loose": [],
				// "sensor_turned": [],
				// "sensor_battery_warning": [],
				// "sensor_over_temperature": [],
				"tire_leak": [],
				"temperature_warning": [],
				"over_inflation_critical": [],
				"over_inflation_warning": [],
				"under_inflation_warning": [],
				"under_inflation_critical": [],
			}
			for (const axle in tires){
				let axle_data = tires[axle]
				for (const tire in axle_data){
					let tire_data = axle_data[tire]
					let tire_ok = true
					tire_data.axle_index = axle.replace("axle_", "")
					tire_data.axle = parseInt(tire_data.axle_index) + 1
					tire_data.tire = tire.replace("tire_", "")
					all_tires.push(tire_data)
					switch (tire_data.sensor_status) {
						case 0:
							tire_ok = false
							tire_signals['sensor_mute'].push(tire_data)
							break
						case 2:
							tire_ok = false
							tire_signals['sensor_defective'].push(tire_data)
							break
					}
					if (tire_data.air_leak_status == 1) {
						tire_ok = false
						tire_signals['tire_leak'].push(tire_data)
					}
					if (tire_data.temperature_state == 1){
						tire_ok = false
						tire_signals['temperature_warning'].push(tire_data)
					}
					switch (tire_data.pressure_state){
						case 0:
							tire_ok = false
							tire_signals['over_inflation_critical'].push(tire_data)
							break
						case 1:
							tire_ok = false
							tire_signals['over_inflation_warning'].push(tire_data)
							break
						case 3:
							tire_ok = false
							tire_signals['under_inflation_warning'].push(tire_data)
							break
						case 4:
							tire_ok = false
							tire_signals['under_inflation_critical'].push(tire_data)
							break
					}
					if (tire_ok){
						tire_signals['alarm_ok'].push(tire_data)
					}
				}
			}
			let tires_update = {}
			for(const signal in tire_signals){
				let tires = tire_signals[signal]
				if (tires.length){
					tires_update[`$tpms.${signal}`] = tires
					signals.push(`@tpms.${signal}`)
				}
			}
			// tires_data
			clear_object(tires_dotted)
			Object.assign(tires_dotted, dotted.dot(tires_update))
			// console.log(tires_dotted)
			Object.assign(syrusjs_variables, tires_dotted);
		}
		fireInternalSignal(...signals)
	}
	function onSleepOn() {
		fireInternalSignal(`@power_save.sleep`);
	}

	// TAIP ex
	// IB:$ibutton.connected.id,IS:$ibutton.last.id
	// IB:$ibutton.authorized.connected.id,IS:$ibutton.authorized.last.id
	// JSON ex
	// connected_ib:$ibutton.connected
	const ib_dotted = {}
	function onIbuttonChange(update){
		let ib_dotted_update = dotted.dot({ $ibutton: update })
		// nullify all keys on ib_dotted
		clear_object(ib_dotted)
		// apply new update
		Object.assign(ib_dotted, ib_dotted_update)
		Object.assign(syrusjs_variables, ib_dotted);
	}

	const rfid_dotted = {}
	function onRFIDMessage(event) {
		// Fire only when rfid changed but not null
		if ( (syrusjs_variables['$rfid.last.id'] != event['last.id']) && (event["last.id"] != null) ) {
			fireInternalSignal('@rfid.change');
		}

		let rfid_dotted_event = dotted.dot({ $rfid: event })
		clear_object(rfid_dotted);
		Object.assign(rfid_dotted, rfid_dotted_event);
		Object.assign(syrusjs_variables, rfid_dotted);

		// Fire event only if rfid lecture generated in the last 5 seconds - current event
		if (syrusjs_variables["$rfid.last.conn_epoch"] >= (Math.floor(Date.now()/1000) - 5) ) {
			// Fire in any rfid event
			fireInternalSignal('@rfid.event');

			// Fire only any card aproached to reader
			if ( syrusjs_variables['$rfid.last.id'] != null   ) {
				fireInternalSignal('@rfid.read');
			}

			// Fire event only when authorized detected
			if ((syrusjs_variables["$rfid.last.whitelisted"] == true) && (syrusjs_variables["$rfid.last.id"] != null)) {
				fireInternalSignal('@rfid.authorized');
			} else if ((syrusjs_variables["$rfid.last.whitelisted"] == false) && (syrusjs_variables["$rfid.last.id"] != null)) {
				fireInternalSignal('@rfid.unauthorized');	
			}
		}	
	}

	function onFuelMessage(event) {
		// Check for Connected / Disconnected transition
		if (syrusjs_variables["$fuel.connected"] == false && event['connected'] == true) {
			fireInternalSignal('@fuel.connected');
		} else if (syrusjs_variables["$fuel.connected"] == true && event['connected'] == false) {
			fireInternalSignal('@fuel.disconnected');
		}

		// State (regular fuel event) / Warning / Fuelling
		if (event['event'] == 'fuelling' || event['event'] == 'fueling') {
			fireInternalSignal('@fuel.fuelling');
			fireInternalSignal('@fuel.fueling');
		} else if (event['event'] == 'warning') {
			fireInternalSignal('@fuel.warning');	
		} else {
			fireInternalSignal('@fuel.state');
		}

		Object.assign(syrusjs_variables, dotted.dot({ $fuel: event}));
		fireInternalSignal('@fuel.event');
	}

	function onWindowEvent(event) {
		

		const name = event.name;
		syrusjs_variables[`$windows.${name}.enabled`] = event.enabled;
		syrusjs_variables[`$windows.${name}.state`] = event.state;
		if (event.state == true) {
			syrusjs_variables[`$windows.${name}.in`] = true;
			syrusjs_variables[`$windows.${name}.out`] = false;
			fireInternalSignal(`@windows.${name}.start`);
		} else {
			syrusjs_variables[`$windows.${name}.in`] = false;
			syrusjs_variables[`$windows.${name}.out`] = true;
			fireInternalSignal(`@windows.${name}.end`);
		}				

	    console.log('onWindowEvent', event, name);
		//console.log('onWindowEvent obj:', syrusjs_variables);
		fireInternalSignal('@windows.event');
	}

	function onSafeEngineEvent(event) {
		syrusjs_variables[`$seco.enabled`] = (event.enabled == 'enable' ) ? true : false;
		syrusjs_variables[`$seco.trigger`] = (event.trigger == 'enable' ) ? true : false;
		syrusjs_variables[`$seco.active`] = (event.trigger == 'enable' ) ? true : false;
		syrusjs_variables[`$seco.state`] = event.state;

		// Add seco activation 
		switch (event.state) {
			case '0':
			    fireInternalSignal('@seco.idle');
		    	break;
			case '1':
			    fireInternalSignal('@seco.canceled');
		    	break;
			case '2':
			    fireInternalSignal('@seco.power_restored', '@seco.deactivated');
		    	break;
			case '3':
			    fireInternalSignal('@seco.power_suspended', '@seco.activated');
		    	break;
			case '4':
			    fireInternalSignal('@seco.activating');
		    	break;	
			case '5':
			    fireInternalSignal('@seco.waiting_no_motion');
		    	break;
			case '6':
			    fireInternalSignal('@seco.waiting_for_gps');
		    	break;
			case '7':
			    fireInternalSignal('@seco.periodic_report');
		    	break;
			default:
				break;																						
		}

		// And state alias
		fireInternalSignal('@seco.event');
	}	

	function onPeopleEvent(event) {

		console.log('onPeopleEvent:', event);
		if (event.event != undefined) {
			const event_type = event.event;
			syrusjs_variables[`$people.event`] = event.event_type;
			Object.assign(syrusjs_variables, dotted.dot({ $people: event}));
			
			// Calculate total adults / children inside
			const total_adults_inside = event.adult.enter - event.adult.exit;
			const total_children_inside = event.children.enter - event.children.exit;
			syrusjs_variables[`$people.adults_inside`] = total_adults_inside;
			syrusjs_variables[`$people.children_inside`] = total_children_inside;
			fireInternalSignal(`@people.${event_type}`)	
		} else {
			// Report type
			const report_type =event.report_type;
			const event_type = `${report_type}_report`;
            syrusjs_variables[`$people.event`] = event.event_type;
			fireInternalSignal(`@people.${event_type}`)	
		}
		
		fireInternalSignal('@people.event');
	}	

	// TAIP ex
	// EA_A:$temp.aliases.sensor1.value
	// EA_A:$temp.sensor_list[0].value
	// JSON ex
	// temp:$temp.aliases
	const temp_dotted = {}
	function onTemperatureChange(update) {
		let temp_dotted_update = dotted.dot({ $temp: update })
		// nullify all keys on ib_dotted
		clear_object(temp_dotted)
		// apply update
		Object.assign(temp_dotted, temp_dotted_update);
		Object.assign(syrusjs_variables, temp_dotted);
	}

	function onSerialMessage(message){
		Object.assign(syrusjs_variables, {"$serial.incoming.message": message});
		fireInternalSignal(`@serial.incoming.message`);
	}

	const fatigue_dotted = {}
	function onFatigueEvent(update){
		let { epoch, event, channel } = update
		if (!epoch) epoch = update.system_epoch;
		if (!event) return
		let recent = (Date.now() / 1e3 - epoch) < 30
		if (!recent || fatigue_dotted['$fatigue_sensor.epoch'] == update.epoch) return
		let fatigue_dotted_update = dotted.dot({ $fatigue_sensor: update })
		// nullify keys on object
		clear_object(fatigue_dotted)
		// apply update
		Object.assign(fatigue_dotted, fatigue_dotted_update)
		Object.assign(syrusjs_variables, fatigue_dotted)
		// deternmine signals
		const signals = [
			`@${channel}.fatigue_sensor.event`,
			`@${channel}.fatigue_sensor.${event}`,
			`@fatigue_sensor.signal`,
			`@fatigue_sensor.${event}`,
			`@dms.signal`,
			`@dms.${event}`,
			`@dms.${channel}.${event}`
		]
		if (fatigue_dotted['$fatigue_sensor.state'] != update.state) {
			signals.push(
				`@${channel}.fatigue_sensor.${update.state}`,
				`@fatigue_sensor.${update.state}`,
			)
		}
		console.log('onFatigueEvent', update);
		if (event.includes('system_error.blurred_image') || event.includes('system_error.over_exposure') ) {
			signals.push(`@dms.${channel}.tamper_alert`)

		}
		if (update.channel.includes('cipia')) {
			signals.push(`@dms.${channel}.event`)
		}	

		// fire signals
		fireInternalSignal(...signals)
	}

	function onMDTMessage(update){
		Object.assign(syrusjs_variables, dotted.dot({ $mdt: update }))
		fireInternalSignal(`@mdt.message`);
	}

	function onLogMessage( event ) {
		Object.assign(syrusjs_variables, dotted.dot({ $logs: event }));
		fireInternalSignal('@logs.event');
	}

	function onBleEvent(event) {
		let deviceList = new Set();
		console.log("onBleEvent", event);
		event.forEach(element => {

			let macAddress = element.mac.substring(0,2) + ':' + element.mac.substring(2,4) + ':' + element.mac.substring(4,6) + ':' + element.mac.substring(6,8) + ':' + element.mac.substring(8,10) + ':' + element.mac.substring(10,12);
			syrusjs_variables[`$ble.${element.mac}.mac`] = macAddress;
			syrusjs_variables[`$ble.${element.mac}.name`] = element.name;
			syrusjs_variables[`$ble.${element.mac}.alias`] = element.name;
			syrusjs_variables[`$ble.${element.mac}.timestamp`] = element.timestamp;

			if (element.alias != undefined || element.alias != null || element.alias != '') {
				syrusjs_variables[`$ble.${element.alias}.mac`] = macAddress;
				syrusjs_variables[`$ble.${element.alias}.name`] = element.name;
				syrusjs_variables[`$ble.${element.alias}.alias`] = element.alias;
				syrusjs_variables[`$ble.${element.alias}.timestamp`] = element.timestamp;
			}

			let tag = `${element.mac}`	
			// Add to detected list
			deviceList.add(element.mac);

			// Check and parse payload
			if (element.payload != undefined || element.payload != null) {
				// Parse characteristic and value
				for (const key in element.payload) {
					// Add the tag 
					tag += `$${key}${element.payload[key]}`;	

					// TODO add more caracteristics to expose to sylang users
					if (key == '2A6E') {
						// Temperature
						let temperature = Number('0x'+element.payload[key], 16);
						
						// Twos complement for negative numbers
						if (element.payload[key][0] == 'F') {
							//console.log('Starting twos complement conversion');
							let byte2 = parseInt('0x'+element.payload[key][1], 16);
							let byte3 = parseInt('0x'+element.payload[key][2], 16);
							let byte4 = parseInt('0x'+element.payload[key][3], 16);
							let tohex = '0x0'+(~byte2*-1).toString(16) +(~byte3*-1).toString(16) +(~byte4*-1).toString(16)
							temperature = -1*parseInt(tohex, 16);
							console.log("Temperature: ", element.payload[key], temperature, Math.round(temperature/100));
						} 

						// Trigger if temp change
						if ( syrusjs_variables[`$ble.${element.mac}.temperature`] !=  Math.round(temperature/100) ) {
							fireInternalSignal(`@ble.${element.mac}.temperature.change`)
						}
						syrusjs_variables[`$ble.${element.mac}.temperature`] = Math.round(temperature/100);
						

						// Add if alias exist
						if (element.alias != undefined || element.alias != null || element.alias != '') { 
							// Trigger if temp change
							if ( syrusjs_variables[`$ble.${element.alias}.temperature`] !=  Math.round(temperature/100) ) {
								fireInternalSignal(`@ble.${element.alias}.temperature.change`)
							}

							syrusjs_variables[`$ble.${element.alias}.temperature`] = Math.round(temperature/100);
						}



					} else if (key == '2A6F') {
						// Humidity
						let humidity = Number('0x'+element.payload[key], 16);
						syrusjs_variables[`$ble.${element.mac}.humidity`] = humidity;

						// Add if alias exist
						if (element.alias != undefined || element.alias != null || element.alias != '') { 
							syrusjs_variables[`$ble.${element.alias}.humidity`] = humidity;
						}

					} else if (key == '2A06') {
						// Movement 
						let movement = Number('0x'+element.payload[key], 16);
						syrusjs_variables[`$ble.${element.mac}.movement`] = movement;

						// Add if alias exist
						if (element.alias != undefined || element.alias != null || element.alias != '') { 
							syrusjs_variables[`$ble.${element.alias}.movement`] = movement;
						}

					} else if (key == '2A3F') {
						// Movement
						// Ignore this
					}
				} 
			}	
			syrusjs_variables[`$ble.${element.mac}.tag`] = tag;
		
		});

		syrusjs_variables[`$ble.list`] = Array.from(deviceList).join(',');
		fireInternalSignal('@ble.event');
	}

	function onEcuWarningEvent(event) {
		Object.assign(syrusjs_variables, dotted.dot({ $ecu_warning: event }));
		console.log("onEcuWarningEvent:", event);
		//syrusjs_variables[`$ecu_warning.${event.id}.${event.warning}`]
		//syrusjs_variables[`$ecu_warning.${event.id}.${event.warning}`]
		fireInternalSignal(`@ecu_warning.${event.id}.${event.warning}`);
		fireInternalSignal('@ecu_warning.event');	
	}	

	function onSystemInfoEvent(event) {
		console.log('systemInfo:', event);

		// Watch out for a reset
		if ( syrusjs_variables[`$system.sessionCount`] != event.sessionCount ) {
			fireInternalSignal('@system.reset');
		}

		// Watch out for network change
		if ( syrusjs_variables[`$system.net`] != event.netLink.name  ) {
			fireInternalSignal('@system.netchange');
		}

		syrusjs_variables[`$system.uptime`] = event.uptime;
		syrusjs_variables[`$system.sessionCount`] = event.sessionCount;
		syrusjs_variables[`$system.net`] = event.netLink.name;
		syrusjs_variables[`$system.ip`] = event.netLink.ip;
	}

	// Establishing all the subcriptions to redis throught the SDK to get the changes on the S4.
	GPS.watchPosition(onGpsData, console.error);
	IOS.watchInputState("*", onInputChange, console.error);
	ECU.watchECUParams(onEcuParamsChange, console.error);
	ECU.onECUWarningEvent(onEcuWarningEvent, console.error);
	Geofences.watchGeofences(onGeofenceChange, console.error, { namespace: configuration.namespace });
	Geofences.watchGroups(onGeofenceGroupChange, console.error, { namespace: configuration.namespace }); 
	Network.onNetworkChange(onNetworkChange, console.error);
	Accelerometer.onMotionChange(onMotionChange, console.error);
	Accelerometer.on(onMotionEvent, console.error);
	System.onSleepOn(onSleepOn, console.error);
	IButton.onIButtonChange(onIbuttonChange, console.error)
	Temperature.onTemperatureChange(onTemperatureChange, console.error)
	Serial.onIncomingMessage(onSerialMessage, console.error)
	Serial.onMDTMessage(onMDTMessage, console.error)
	SafeEngine.onSafeEngineEvent(onSafeEngineEvent, console.error)
	RFID.onRFIDEvent(onRFIDMessage, console.error)
	Technoton.onFuelEvent(onFuelMessage, error => console.log('technoton error', error));
	Logrotate.onLogEvent(onLogMessage, console.error);
	Fatigue.onFatigueEvent(onFatigueEvent, console.error);
	Windows.onWindowEvent(onWindowEvent, console.error);
	PeopleCounter.onPeopleCountingEvent(onPeopleEvent, console.error);
	Bluetooth.onBleUpdate(onBleEvent, console.error);
	Bluetooth.onBluetoothUpdate(
		(channel, payload) => {
			// console.log('bt.watcher', {channel, payload})
			// backwards compatability
			if (channel.endsWith('/CALL_BUTTON')){
				fireInternalSignal(
					"@bluetooth.button.signal",
					`@bluetooth.button.call_${payload}x`,
				)
				return
			}
			if (channel.endsWith('/action')){
				let { button, action, action_multiplier } = payload
				if (action != 'tap') return
				fireInternalSignal(
					"@bluetooth.button.signal",
					`@bluetooth.button.${button}_${action_multiplier}x`,
				)
				return
			}
			if (channel.endsWith('/update')) {
				console.log('onBluetoothUpdate:', payload);
				Object.assign(syrusjs_variables, dotted.dot({ $bluetooth: payload}));
				fireInternalSignal('@bluetooth.event');

				if (payload.connected === true) {
					syrusjs_variables[`$bluetooth.state`] = "connected";
					fireInternalSignal('@bluetooth.connected');
				} else {
					syrusjs_variables[`$bluetooth.state`] = "disconnected";
					fireInternalSignal('@bluetooth.disconnected');
				}	

				if (payload.paired === true) {
					fireInternalSignal('@bluetooth.paired');
				} else {
					fireInternalSignal('@bluetooth.unpaired');
				}
				
			}

		},
		console.error
	)

	// Get the current state of network, ios and ecu, to update the signals and variables
	Network.getNetworks().then(onNetworkChange).catch(console.error);
	Geofences.getAll({ namespace: configuration.namespace }).then(onGeofenceChange).catch(console.error);
	IOS.getAll().then(onInputChange).catch(console.error);
	ECU.getECUParams().then(onEcuParamsChange).catch(console.error);
	System.info().then(onSystemInfoEvent).catch(console.error);

	/**
	 * All the config file are watch in case of any change to restart the app a get the new configuration
	 * TODO: pending improve  get the new config without a restart.
	 */
	var { destinationsPath, configPath, schemaPath } = configuration;
	fs.watch(schemaPath, () => {
		console.log(`detected changes on file ${schemaPath}, restarting app`);
		process.kill(process.pid, "SIGTERM");
	});
	fs.watch(destinationsPath, () => {
		console.log(`detected changes on file ${destinationsPath}, restarting app`);
		process.kill(process.pid, "SIGTERM");
	});
	fs.watch(configPath, () => {
		console.log(`detected changes on file ${configPath}, restarting app`);
		process.kill(process.pid, "SIGTERM");
	});

	/* Arbitrary callback to check if syrus was restarted by psm.
	This is not called immediately since the unix time can be invalid while there is no network
	and it is established to be the internal clock of S4
	*/
	setTimeout(async () => {
		try{
			let wakeup = await System.getLastWakeUp()
			if (wakeup) {
				const newTimestamp = dayjs(wakeup.timestamp)
				let reason = wakeup.wakeup_reason.toLowerCase()
				if (reason == "mot") reason = "motion"
				if (reason == "elapsed_time") reason = "time"
				if (reason == "electrical_rst") reason = "electrical_reset"
				if (newTimestamp.unix() != syrusjs_variables["$power_save.time"]) {
					syrusjs_variables["$power_save.time"] = newTimestamp.unix()
					syrusjs_variables["$power_save.reason"] = wakeup.wakeup_reason
					fireInternalSignal(
						`@power_save.wakeup`,
						`@power_save.${reason}`,
					)
				}
			}
		} catch(err){
			console.error("ERROR getting wakeup reason")
			console.error(err)
		}
	}, 1000 * 15);

	// Interval for sync data from the utils within syrusjs
	// In case redis fails for some reason or the syrus-interface fails for any reason
	// Could be improved this handling this part inside the Syrus4-SDK
	setInterval(() => {
		onNetworkChange();
		IOS.getAll().then(onInputChange).catch(console.error);
		ECU.getECUParams().then(onEcuParamsChange).catch(console.error);
	}, 1000 * 60 * 5);

	// update $modem information periodically
	// TODO: update with publish
	setInterval(async ()=>{
		try {
			var modem = await System.modem()
			for (const key in modem) {
				syrusjs_variables[`$modem.${key.toLowerCase()}`] = modem[key]
			}
		} catch (error) {
			console.error(error)
		}
	}, 1000 * 60 * 1)

	// Update 
};
